-- Class: Engine (3D)
-- SDK: Corona - 09/12/11
-- Author: Andrew Burch
-- Date: 02/03/12
-- Site: http://www.newhighscore.net
-- Contact: andrew.burch@newhighscore.net
-- Note: 3D Engine handles rotation and sorting of points and polys

local tablex = require('core.tablex')

local cos = math.cos
local sin = math.sin
local rad = math.rad
local floor = math.floor


local Engine = {}
local Engine_mt = {__index = Engine}

function Engine.new(params)
	local class = {
		objectList = {},
	}
	
	return setmetatable(class, Engine_mt)
end

function Engine:initialise()
	local sinTable = {}
	local cosTable = {}
	for i = 0, 360 do 
		sinTable[i] = sin(rad(i))
		cosTable[i] = cos(rad(i))
	end

	self.sinTable = sinTable
	self.cosTable = cosTable
end

function Engine:registerObject(params)
	local id = params.id
	local object = params.object
	local rotationSequence = params.rotationSequence
	
	self.objectList[id] = {
					object = object,
					rotationSequence = rotationSequence,
				}
end

function Engine:getObject(id)
	local objectList = self.objectList
	local registeredObject = objectList[id]
	
	if registeredObject then
		return registeredObject.object
	end
end

function Engine:update(dt, time)
	local objectList = self.objectList

	for _, objectInfo in pairs(objectList) do
		local object = objectInfo.object
		local rotationSequence = objectInfo.rotationSequence

		-- rotate points
		local vertList = object:getVertList()
		local rotation = object:getRotation()
		local rotatedVertList = self:rotateVertList(vertList, rotation, rotationSequence)
		
		-- rotate face normals
		local faceNormals = object:getFaceNormals()
		local rotatedNormals = self:rotateVertList(faceNormals, rotation, rotationSequence)
		
		-- sort polys
		local polyList = object:getPolyList()
		local sortedPolys = self:sortPolys(polyList, rotatedVertList, rotatedNormals)

		-- update object
		object:setRotatedVertList(rotatedVertList)
		object:setRotatedFaceNormals(rotatedNormals)
		object:setSortedPolyList(sortedPolys)
		
		-- update axis rotation of the object
		for _, rotInfo in pairs(rotationSequence) do
			local axis = rotInfo.axis
			local angleStep = rotInfo.angleStep
			
			object:updateRotationAngle(axis, angleStep)
		end
	end
end

function Engine:rotateVertList(vertList, rotation, rotationSequence)
	local sinTable = self.sinTable
	local cosTable = self.cosTable
	
	local output = {}
	
	for i, vertSource in ipairs(vertList) do
		local vert = {
			x = vertSource.x,
			y = vertSource.y,
			z = vertSource.z,
		}

		for _, rotInfo in ipairs(rotationSequence) do
			local axis = rotInfo.axis
			local angle = floor(rotation[axis])
			
			if axis == 'x' then
				local yt = cosTable[angle] * vert.y - sinTable[angle] * vert.z 
				local zt = sinTable[angle] * vert.y + cosTable[angle] * vert.z
				vert.y = yt
				vert.z = zt
			end
		
			if axis == 'y' then
				local xt = cosTable[angle] * vert.x - sinTable[angle] * vert.z
				local zt = sinTable[angle] * vert.x + cosTable[angle] * vert.z
				vert.x = xt
				vert.z = zt
			end
			
			if axis == 'z' then
				local xt = cosTable[angle] * vert.x - sinTable[angle] * vert.y
				local yt = sinTable[angle] * vert.x + cosTable[angle] * vert.y
				vert.x = xt
				vert.y = yt
			end
		end
		
		table.insert(output, vert)
	end

	return output
end

function Engine:sortPolys(polyList, vertList, normalList)
	local output = {}
	
	for i, polyInfo in ipairs(polyList) do
		local polyVerts = polyInfo.verts
		local z = 0
		local vertTotal = 0
		for _, vindex in ipairs(polyVerts) do
			z = z + vertList[vindex].z
			vertTotal = vertTotal + 1
		end
		
		local averageZ = z / vertTotal
				
		table.insert(output, {
						id = polyInfo.id,
						verts = polyVerts,
						normal = normalList[i],
						z = averageZ,
					})
	end
	
	table.sort(output, function(a, b) return a.z < b.z end)
	
	return output
end	


return Engine