-- Class: 3D Object Renderer
-- SDK: Corona - 09/12/11
-- Author: Andrew Burch
-- Date: 02/03/12
-- Site: http://www.newhighscore.net
-- Contact: andrew.burch@newhighscore.net
-- Note: Renders a 3D object to screen using various methods

local vector = require('math.vector')

local round = math.round
local floor = math.floor


local Renderer = {}
local Renderer_mt = {__index = Renderer}

function Renderer.new(params)
	local displayGroup = display.newGroup()
	
	local class = {
		displayGroup = displayGroup,
	}
	
	return setmetatable(class, Renderer_mt)
end

function Renderer:initialise(params)
	local displayGroup = self.displayGroup
	
	-- used for vertex only rendering
	local vertObjects = {}
	local maxVerts = 8
	for i = 1, maxVerts do
		local circle = display.newCircle(displayGroup, 0, 0, 3)
		circle:setFillColor(200, 200, 200)
		circle.alpha = 0
		table.insert(vertObjects, circle)
	end
	
	-- used for rendering polygons
	local polyLines = {}
	local maxLines = 580
	for i = 1, maxLines do
		local slice = display.newRect(displayGroup, 0, 0, 4, 1)
		slice:setFillColor(0, 0, 0)
		slice.alpha = 0
		
		table.insert(polyLines, slice)
	end
	
	self.maxVerts = maxVerts
	self.maxLines = maxLines
	self.polyLines = polyLines
	self.vertObjects = vertObjects
	self.wireLines = {}
	self.maxWireLines = 24
end

function Renderer:reset()
print('reset')
	-- reset poly lines
	local polyLines = self.polyLines
	local lastUsedLineCount = self.lastUsedLineCount or 0
	for i = 1, lastUsedLineCount do
		local line = polyLines[i]
		line.alpha = 0
	end
	
	-- reset vert objects
	local vertObjects = self.vertObjects
	for _, v in ipairs(vertObjects) do
		v.alpha = 0
	end
	
	-- clear wireframe lines
	local wireLines = self.wireLines
	local displayGroup = self.displayGroup
	for _, v in ipairs(wireLines) do
		displayGroup:remove(v)
	end
	
	self.wireLines = {}
end

function Renderer:renderObject(params)
	local mode = params.renderMode

	if mode == 'wire' then	
		self:renderObjectWireframe(params)
	elseif mode == 'vert' then
		self:renderObjectVerts(params)
	else
		self:renderObjectPolys(params)
	end
end

function Renderer:renderObjectPolys(params)
	local polyLines = self.polyLines
	
	local displayWidth = params.displayWidth
	local displayHeight = params.displayHeight
	local object = params.object
	local useLightMap = params.lightMap
	local lightVector = params.lightVector
	local alpha = params.polyAlpha

	local renderPolys = self:buildRenderPolys(params)

	local lineIndex = 1
	
	for _, poly in ipairs(renderPolys) do
		local verts = poly.vertList
		local miny = 1000
		local maxy = -1000
		for _, vert in ipairs(verts) do
			miny = vert.y < miny and vert.y or miny
			maxy = vert.y > maxy and vert.y or maxy
		end
		
		local xlist = {}
		self:scanEdge(xlist, displayWidth, verts[1].x, verts[1].y, verts[2].x, verts[2].y)
		self:scanEdge(xlist, displayWidth, verts[2].x, verts[2].y, verts[3].x, verts[3].y)
		self:scanEdge(xlist, displayWidth, verts[3].x, verts[3].y, verts[4].x, verts[4].y)
		self:scanEdge(xlist, displayWidth, verts[4].x, verts[4].y, verts[1].x, verts[1].y)

		local faceColour = poly.faceColour
		local red = faceColour.r
		local green = faceColour.g
		local blue = faceColour.b
		if useLightMap then
			local normal = poly.normal
		    local lightFactor = (normal.x * lightVector.x) + (normal.y * lightVector.y) + (normal.z * lightVector.z)
		    red = red * lightFactor
		    green = green * lightFactor
		    blue = blue * lightFactor
		end

		for y = miny, maxy do
			local line = polyLines[lineIndex]
			local lineLength = xlist[y].b - xlist[y].a
			if lineLength > 0 then
				local x = xlist[y].a + (lineLength  * 0.5)
				line:toFront()
				line:setFillColor(red, green , blue)
				line.x = x
				line.y = y
				line.width = lineLength
				line.alpha = alpha
				
				lineIndex = lineIndex + 1
			end
		end
	end
	
	local linesUsed = lineIndex - 1
	local lastUsedLineCount = self.lastUsedLineCount or 0
	if linesUsed < lastUsedLineCount then
		for i = lineIndex, lastUsedLineCount do
			local line = polyLines[i + 1]
			line.alpha = 0
		end
	end

	self.lastUsedLineCount = lineIndex	
end

function Renderer:scanEdge(xlist, displayWidth, x1, y1, x2, y2)
	if y1 == y2 then
		return 
	end
	
	if y1 > y2 then
		local ty = y1
		local tx = x1
		y1 = y2
		y2 = ty
		x1 = x2
		x2 = tx
	end
	local xstep = ((x2 - x1) * 256) / (y2 - y1)
	local x = x1 * 256
	for i = y1, y2 do
		local xfloor = floor(x / 256)
		if not xlist[i] then
			xlist[i] = {a = displayWidth, b = 0}
		end
		
		if xfloor < xlist[i].a then 
			xlist[i].a = xfloor
		end
		
		if xfloor > xlist[i].b then
			xlist[i].b = xfloor
		end
		
		x = x + xstep
	end
end

function Renderer:buildRenderPolys(params)
	local object = params.object
	local faceColours = params.faceColours

	local polyList = object:getSortedPolyList()
	local vertList = object:getRotatedVertList()
	local faceNormals = object:getRotatedFaceNormals()

	local xoffset = object:getAxisOffset('x')
	local yoffset = object:getAxisOffset('y')
	local zoffset = object:getAxisOffset('z')
	
	local removeHiddenSurface = params.removeHiddenSurface

	local renderPolys = {}
	for i, poly in ipairs(polyList) do
		local transformedVerts = {}

		for _, vertIndex in ipairs(poly.verts) do
			local vert = vertList[vertIndex]
			local z = vert.z + zoffset
			local x = floor((256 * vert.x) / z + xoffset)
			local y = floor((256 * vert.y) / z + yoffset)
			
			table.insert(transformedVerts, {
								x = x,
								y = y,
								z = z,
							})
		end
		
		local includeFace = true
		if removeHiddenSurface then
			local faceNormal = vector.faceZNormal3(transformedVerts)
	    	includeFace = faceNormal > 0
	    end
	    
	    if includeFace then
	    	table.insert(renderPolys, { 
				    			vertList = transformedVerts, 
				    			faceColour = faceColours[poly.id], 
				    			normal = poly.normal,
				    		})
	  	end
	end
	
	return renderPolys
end

function Renderer:renderObjectWireframe(params)
	local displayGroup = self.displayGroup

	local wireLines = self.wireLines
	local displayGroup = self.displayGroup
	for _, v in ipairs(wireLines) do
		displayGroup:remove(v)
	end	
	self.wireLines = {}
	
	local renderPolys = self:buildRenderPolys(params)	
	
	for _, v in ipairs(renderPolys) do
		local vertList = v.vertList
		local vertCount = #vertList
		for i = 1, vertCount do
			local v1 = i
			local v2 = i == vertCount and 1 or i + 1
			local x1 = vertList[v1].x
			local y1 = vertList[v1].y
			local x2 = vertList[v2].x
			local y2 = vertList[v2].y
			
			local line = display.newLine(displayGroup, x1, y1, x2, y2)
			line:setColor(200, 200, 200)
			line.strokeWidth = 2
			
			table.insert(self.wireLines, line)
		end
	end
end

function Renderer:renderObjectVerts(params)
	local vertObjects = self.vertObjects

	local displayWidth = params.displayWidth
	local displayHeight = params.displayHeight
	local object = params.object

	local xoffset = object:getAxisOffset('x')
	local yoffset = object:getAxisOffset('y')
	local zoffset = object:getAxisOffset('z')

	local rotatedVerts = object:getRotatedVertList()
	
	for i, vert in ipairs(rotatedVerts) do
		local z = vert.z + zoffset
		local zClipped = z >= 0
		local renderObject = vertObjects[i]
		renderObject.alpha = zClipped and 0 or 1
		
		if not zClipped then
			local temp = floor(z)
			local nx = floor(256 * vert.x / temp) + xoffset
			local ny = floor(256 * vert.y / temp) + yoffset

			local xClipped = nx < 0 or nx > displayWidth
			local yClipped = ny < 0 or ny > displayHeight
			local clipped = xClipped or yClipped

			renderObject.alpha = clipped and 0 or 1
			
			if not clipped then
				renderObject.x = nx
				renderObject.y = ny
			end			
		end
	end	
end

return Renderer